import ctypes
from .enums import (CorsairEventId, CorsairKeyId, CorsairLedId,
                    CorsairDeviceType, CorsairChannelDeviceType,
                    CorsairPhysicalLayout, CorsairLogicalLayout)
from .utils import bytes_to_str_or_default

__all__ = [
    'CorsairProtocolDetails', 'CorsairChannelDeviceInfo', 'CorsairChannelInfo',
    'CorsairChannelsInfo', 'CorsairDeviceInfo', 'CorsairLedPosition',
    'CorsairLedPositions', 'CorsairLedColor', 'CorsairEvent',
    'CorsairKeyEvent', 'CorsairDeviceConnectionStatusChangedEvent'
]

CORSAIR_DEVICE_ID_MAX = 128


class CorsairProtocolDetails(ctypes.Structure):
    _fields_ = [('sdkVersion', ctypes.c_char_p),
                ('serverVersion', ctypes.c_char_p),
                ('sdkProtocolVersion', ctypes.c_int32),
                ('serverProtocolVersion', ctypes.c_int32),
                ('breakingChanges', ctypes.c_bool)]

    def __str__(self):
        return ("\n".join([
            "SDK version: " + bytes_to_str_or_default(self.sdkVersion),
            "SDK protocol version: " + str(self.sdkProtocolVersion),
            "Server version: " + bytes_to_str_or_default(self.serverVersion),
            "Server protocol version: " + str(self.serverProtocolVersion),
            "Breaking changes: " + str(self.breakingChanges)
        ]))


class CorsairChannelDeviceInfo(ctypes.Structure):
    _fields_ = [('type', CorsairChannelDeviceType),
                ('deviceLedCount', ctypes.c_int32)]


class CorsairChannelInfo(ctypes.Structure):
    _fields_ = [('totalLedsCount', ctypes.c_int32),
                ('devicesCount', ctypes.c_int32),
                ('devices', ctypes.POINTER(CorsairChannelDeviceInfo))]


class CorsairChannelsInfo(ctypes.Structure):
    _fields_ = [('channelsCount', ctypes.c_int32),
                ('channels', ctypes.POINTER(CorsairChannelInfo))]


class CorsairDeviceInfo(ctypes.Structure):
    _fields_ = [('type', CorsairDeviceType), ('model', ctypes.c_char_p),
                ('physicalLayout', CorsairPhysicalLayout),
                ('logicalLayout', CorsairLogicalLayout),
                ('capsMask', ctypes.c_int32), ('ledsCount', ctypes.c_int32),
                ('channels', CorsairChannelsInfo),
                ('deviceId', ctypes.c_char * CORSAIR_DEVICE_ID_MAX)]


class CorsairLedPosition(ctypes.Structure):
    _fields_ = [('ledId', CorsairLedId), ('top', ctypes.c_double),
                ('left', ctypes.c_double), ('height', ctypes.c_double),
                ('width', ctypes.c_double)]


class CorsairLedPositions(ctypes.Structure):
    _fields_ = [('numberOfLed', ctypes.c_int32),
                ('pLedPosition', ctypes.POINTER(CorsairLedPosition))]


class CorsairLedColor(ctypes.Structure):
    _fields_ = [('ledId', CorsairLedId), ('r', ctypes.c_int32),
                ('g', ctypes.c_int32), ('b', ctypes.c_int32)]


class CorsairDeviceConnectionStatusChangedEvent(ctypes.Structure):
    _fields_ = [('deviceId', ctypes.c_char * CORSAIR_DEVICE_ID_MAX),
                ('isConnected', ctypes.c_bool)]


class CorsairKeyEvent(ctypes.Structure):
    _fields_ = [('deviceId', ctypes.c_char * CORSAIR_DEVICE_ID_MAX),
                ('keyId', CorsairKeyId), ('isPressed', ctypes.c_bool)]


class CorsairEventPayload(ctypes.Union):
    _fields_ = [('deviceConnectionStatusChangedEvent',
                 ctypes.POINTER(CorsairDeviceConnectionStatusChangedEvent)),
                ('keyEvent', ctypes.POINTER(CorsairKeyEvent))]


class CorsairEvent(ctypes.Structure):
    _anonymous_ = ("payload", )
    _fields_ = [('id', CorsairEventId), ('payload', CorsairEventPayload)]



import os
import platform
import sys
from ctypes import (CDLL, CFUNCTYPE, POINTER, sizeof, c_bool, c_char, c_int32,
                    c_void_p)
from .enums import (CorsairAccessMode, CorsairError, CorsairLedId,
                    CorsairEventId, CorsairDevicePropertyId)
from .structs import (CorsairProtocolDetails, CorsairDeviceInfo,
                      CorsairLedPosition, CorsairLedPositions, CorsairLedColor,
                      CorsairEvent)

__all__ = ['CorsairNativeApi']


def get_library_path_windows():
    suffix = '.x64' if sizeof(c_void_p) == 8 else ''
    lib_name = 'CUESDK' + suffix + '_2017.dll'
    return os.path.join(os.path.dirname(__file__), 'bin', lib_name)


def get_library_path_mac():
    lib_name = 'libCUESDK.dylib'
    return os.path.join(os.path.dirname(__file__), 'bin', lib_name)


def load_library(library_path):
    try:
        return CDLL(library_path)
    except OSError:
        print("Unable to load the library %s" % library_path)
        sys.exit()


class CorsairNativeApi():
    def __init__(self, libpath):
        if libpath is None:
            system = platform.system()
            if system == "Windows":
                libpath = get_library_path_windows()
            elif system == "Darwin":
                libpath = get_library_path_mac()
        lib = load_library(libpath)

        def create_func(fn, restype, argtypes):
            f = lib.__getattr__(fn)
            f.restype = restype
            f.argtypes = argtypes
            return f

        self.CorsairSetLedsColorsBufferByDeviceIndex = create_func(
            'CorsairSetLedsColorsBufferByDeviceIndex', c_bool,
            [c_int32, c_int32, POINTER(CorsairLedColor)])
        self.CorsairSetLedsColorsFlushBuffer = create_func(
            'CorsairSetLedsColorsFlushBuffer', c_bool, None)
        self.CallbackFunc = CFUNCTYPE(c_void_p, c_bool, CorsairError)
        self.CorsairSetLedsColorsFlushBufferAsync = create_func(
            'CorsairSetLedsColorsFlushBufferAsync', c_bool,
            [self.CallbackFunc, c_void_p])
        self.CorsairGetLedsColors = create_func(
            'CorsairGetLedsColors', c_bool,
            [c_int32, POINTER(CorsairLedColor)])
        self.CorsairGetLedsColorsByDeviceIndex = create_func(
            'CorsairGetLedsColorsByDeviceIndex', c_bool,
            [c_int32, c_int32, POINTER(CorsairLedColor)])
        self.CorsairGetDeviceCount = create_func('CorsairGetDeviceCount',
                                                 c_int32, None)
        self.CorsairGetDeviceInfo = create_func('CorsairGetDeviceInfo',
                                                POINTER(CorsairDeviceInfo),
                                                [c_int32])
        self.CorsairGetLedPositions = create_func('CorsairGetLedPositions',
                                                  POINTER(CorsairLedPositions),
                                                  None)
        self.CorsairGetLedPositionsByDeviceIndex = create_func(
            'CorsairGetLedPositionsByDeviceIndex',
            POINTER(CorsairLedPositions), [c_int32])
        self.CorsairGetLedIdForKeyName = create_func(
            'CorsairGetLedIdForKeyName', CorsairLedId, [c_char])
        self.CorsairRequestControl = create_func('CorsairRequestControl',
                                                 c_bool, [CorsairAccessMode])
        self.CorsairPerformProtocolHandshake = create_func(
            'CorsairPerformProtocolHandshake', CorsairProtocolDetails, None)
        self.CorsairGetLastError = create_func('CorsairGetLastError',
                                               CorsairError, None)
        self.CorsairReleaseControl = create_func('CorsairReleaseControl',
                                                 c_bool, [CorsairAccessMode])
        self.CorsairSetLayerPriority = create_func('CorsairSetLayerPriority',
                                                   c_bool, [c_int32])
        c_bool_p = POINTER(c_bool)
        self.CorsairGetBoolPropertyValue = create_func(
            'CorsairGetBoolPropertyValue', c_bool,
            [c_int32, CorsairDevicePropertyId, c_bool_p])
        c_int32_p = POINTER(c_int32)
        self.CorsairGetInt32PropertyValue = create_func(
            'CorsairGetInt32PropertyValue', c_bool,
            [c_int32, CorsairDevicePropertyId, c_int32_p])
        self.EventHandler = CFUNCTYPE(None, c_void_p, POINTER(CorsairEvent))
        self.CorsairSubscribeForEvents = create_func(
            'CorsairSubscribeForEvents', c_bool, [self.EventHandler, c_void_p])
        self.CorsairUnsubscribeFromEvents = create_func(
            'CorsairUnsubscribeFromEvents', c_bool, None)



from ctypes import c_bool, c_int32, byref
from typing import Any, Optional, Union, Callable, Dict, Sequence, Tuple

from .capi import CorsairNativeApi
from .enums import (CorsairAccessMode, CorsairError, CorsairDevicePropertyId,
                    CorsairDevicePropertyType, CorsairLedId, CorsairEventId)
from .helpers import ColorRgb
from .structs import (CorsairProtocolDetails, CorsairDeviceInfo,
                      CorsairLedPosition, CorsairLedPositions, CorsairLedColor,
                      CorsairEvent)
from .utils import bytes_to_str_or_default

__all__ = ['CueSdk']

native = None


class Device(object):
    def __init__(self, id, type, model, caps_mask, led_count, channels):
        self.id = id
        self.type = type
        self.model = model
        self.caps_mask = caps_mask
        self.led_count = led_count
        self.channels = channels

    def __repr__(self):
        return self.model


class Channel(object):
    def __init__(self, total_led_count, devices):
        self.total_led_count = total_led_count
        self.devices = devices


class ChannelDevice(object):
    def __init__(self, type, led_count):
        self.type = type
        self.led_count = led_count


class CueSdk(object):
    def __init__(self, sdk_path: Optional[str] = None) -> None:
        global native
        native = CorsairNativeApi(sdk_path)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def connect(self):
        self.protocol_details = native.CorsairPerformProtocolHandshake()
        err = native.CorsairGetLastError()
        return err == CorsairError.Success

    def get_devices(self):
        cnt = native.CorsairGetDeviceCount()
        if cnt > 0:
            return [self.get_device_info(i) for i in range(cnt)]
        return []

    def get_device_count(self) -> int:
        """Returns number of connected Corsair devices.

        For keyboards, mice, mousemats, headsets and headset stands not more
        than one device of each type is included in return value in case if
        there are multiple devices of same type connected to the system.
        For DIY-devices and coolers actual number of connected devices is
        included in return value. For memory modules actual number of
        connected modules is included in return value, modules are enumerated
        with respect to their logical position (counting from left to right,
        from top to bottom).

        Use `CueSdk.get_device_info` to get information about a certain device.
        """
        return native.CorsairGetDeviceCount()

    def get_device_info(self, device_index: int):
        """Returns information about a device based on provided index.

        Args:
            device_index: zero-based index of device. Should be strictly
                less than a value returned by `CueSdk.get_device_count`
        """
        p = native.CorsairGetDeviceInfo(device_index)
        s = p.contents
        channels = []
        for chi in range(s.channels.channelsCount):
            ch = s.channels.channels[chi]
            devices = [
                ChannelDevice(ch.devices[i].type, ch.devices[i].deviceLedCount)
                for i in range(ch.devicesCount)
            ]
            channels.append(Channel(ch.totalLedsCount, devices))
        return Device(bytes_to_str_or_default(s.deviceId), s.type,
                      bytes_to_str_or_default(s.model), s.capsMask,
                      s.ledsCount, channels)

    def get_last_error(self) -> CorsairError:
        """Returns last error that occurred in this thread while using
        any of SDK functions.
        """
        return native.CorsairGetLastError()

    def request_control(self) -> bool:
        """Requests exclusive control over lighting.

        By default client has shared control over lighting so there is
        no need to call `CueSdk.request_control` unless a client
        requires exclusive control.
        """
        return native.CorsairRequestControl(
            CorsairAccessMode.ExclusiveLightingControl)

    def release_control(self) -> bool:
        """Releases previously requested exclusive control."""
        return native.CorsairReleaseControl(
            CorsairAccessMode.ExclusiveLightingControl)

    def set_layer_priority(self, priority: int) -> bool:
        """Sets layer priority for this shared client.

        By default CUE has priority of 127 and all shared clients have
        priority of 128 if they don't call this function. Layers with
        higher priority value are shown on top of layers with lower priority.

        Args:
            priority: priority of a layer [0..255]

        Returns:
            boolean value. True if successful. Use `CueSdk.get_last_error`
            to check the reason of failure. If this function is called in
            exclusive mode then it will return True
        """
        return native.CorsairSetLayerPriority(priority)

    def get_led_id_for_key_name(self, key_name: str) -> CorsairLedId:
        """Retrieves led id for key name taking logical layout into account.

        So on AZERTY keyboards if user calls `get_led_id_for_key_name('A')` he
        gets `CorsairLedId.K_Q`. This id can be used in
        `CueSdk.set_led_colors_buffer_by_device_index` function

        Args:
            key_name: key name. ['A'..'Z'] (26 values) are valid values

        Returns:
            Proper `CorsairLedId` or `CorsairLedId.Invalid` if error
            occurred
        """
        if not isinstance(key_name,
                          str) or len(key_name) != 1 or not key_name.isalpha():
            return CorsairLedId(CorsairLedId.Invalid)

        return native.CorsairGetLedIdForKeyName(key_name.encode())

    def set_led_colors_buffer_by_device_index(
            self, device_index: int,
            led_colors: Dict[CorsairLedId, Tuple[int, int, int]]) -> bool:
        """Sets specified LEDs to some colors.

        This function set LEDs colors in the buffer which is written to the
        devices via `CueSdk.set_led_colors_flush_buffer` or
        `CueSdk.set_led_colors_flush_buffer_async`.
        Typical usecase is next: `CueSdk.set_led_colors_flush_buffer` or
        `CueSdk.set_led_colors_flush_buffer_async` is called to write LEDs
        colors to the device and follows after one or more calls of
        `CueSdk.set_led_colors_buffer_by_device_index` to set the LEDs buffer.
        This function does not take logical layout into account.

        Args:
            device_index: zero-based index of device. Should be strictly less
                than value returned by `CueSdk.get_device_count`
            led_colors: a dict mapping `CorsairLedId` keys to the corresponding
                color values. Each color is represented as a tuple of ints
                (r, g, b). For example:

            ```python
            { CorsairLedId.K_F1: (255, 0, 0) }
            ```
        """
        sz = len(led_colors)
        data = (CorsairLedColor * sz)()
        rgb_keys = ['r', 'g', 'b']
        for i, led in enumerate(led_colors):
            rgb = dict(zip(rgb_keys, led_colors[led]))
            data[i] = CorsairLedColor(ledId=led, **rgb)
        return native.CorsairSetLedsColorsBufferByDeviceIndex(
            device_index, sz, data)

    def set_led_colors_flush_buffer(self) -> bool:
        """Writes to the devices LEDs colors buffer which is previously filled
        by the `CueSdk.set_led_colors_buffer_by_device_index` function.

        This function executes synchronously, if you are concerned about
        delays consider using `CueSdk.set_led_colors_flush_buffer_async`
        """
        return native.CorsairSetLedsColorsFlushBuffer()

    def set_led_colors_flush_buffer_async(self) -> bool:
        """Same as `CueSdk.set_led_colors_flush_buffer` but returns control to
        the caller immediately
        """
        return native.CorsairSetLedsColorsFlushBufferAsync(None, None)

    def get_led_colors_by_device_index(
        self, device_index: int, led_identifiers: Sequence[CorsairLedId]
    ) -> Union[None, Dict[CorsairLedId, Tuple[int, int, int]]]:
        """Gets current color for the requested LEDs.

        The color should represent the actual state of the hardware LED, which
        could be a combination of SDK and/or CUE input. This function works
        for keyboard, mouse, mousemat, headset, headset stand, DIY-devices,
        memory module and cooler.

        Args:
            device_index: zero-based index of device. Should be strictly less
                than value returned by `CueSdk.get_device_count`
            led_identifiers: the list of `CorsairLedId` values

        Returns:
            A dict mapping `CorsairLedId` keys to the corresponding color
            values. Each color is represented as a tuple of ints. For
            example:

            ```python
            {
              CorsairLedId.K_Escape: (255, 0, 0),
              CorsairLedId.K_F1: (128, 0, 128)
            }
            ```

            If error occurred, the function returns None
        """
        sz = len(led_identifiers)
        data = (CorsairLedColor * sz)()
        for i in range(sz):
            data[i].ledId = led_identifiers[i]
        ok = native.CorsairGetLedsColorsByDeviceIndex(device_index, sz, data)
        if not ok:
            return None
        ret = {}
        for i in range(sz):
            c = data[i]
            ret[c.ledId] = (c.r, c.g, c.b)
        return ret

    def get_led_positions_by_device_index(
        self, device_index: int
    ) -> Union[None, Dict[CorsairLedId, Tuple[float, float]]]:
        """Provides dictionary of keyboard, mouse, headset, mousemat,
        headset stand, DIY-devices, memory module and cooler LEDs by its index
        with their positions.

        Position could be either physical (only device-dependent)
        or logical (depend on device as well as CUE settings).

        Args:
            device_index: zero-based index of device. Should be strictly less
                than value returned by `CueSdk.get_device_count`

        Returns:
            A dict mapping `CorsairLedId` keys to the corresponding LED
            positions. Each position is represented as a tuple of floats.
            For example:

            ```python
            { CorsairLedId.K_Escape: (77.0, 36.0) }
            ```

            If error occurred, the function returns None
        """
        leds = native.CorsairGetLedPositionsByDeviceIndex(device_index)
        if not leds:
            return None
        positions = {}
        for i in range(leds.contents.numberOfLed):
            p = leds.contents.pLedPosition[i]
            positions[p.ledId] = (p.left, p.top)
        return positions

    def subscribe_for_events(
            self, handler: Callable[[CorsairEventId, Any], None]) -> bool:
        """Registers a callback that will be called by SDK when some event
        happened.

        If client is already subscribed but calls this function again SDK
        should use only last callback registered for sending notifications.

        Args:
            handler:
                Callback function with two arguments: event_id and event_data

        Returns:
            boolean value. True if successful. Use `CueSdk.get_last_error`
            to check the reason of failure.
        """
        if handler is None:
            return False

        def raw_handler(ctx, e):
            id = e[0].id
            if (id == CorsairEventId.DeviceConnectionStatusChangedEvent):
                handler(id, e[0].deviceConnectionStatusChangedEvent[0])
            elif (id == CorsairEventId.KeyEvent):
                handler(id, e[0].keyEvent[0])

        self.event_handler = native.EventHandler(raw_handler)
        return native.CorsairSubscribeForEvents(self.event_handler, None)

    def unsubscribe_from_events(self):
        """Unregisters callback previously registered by
        `CueSdk.subscribe_for_events` call

        Returns:
            boolean value. True if successful. Use `CueSdk.get_last_error`
            to check the reason of failure.
        """
        self.event_handler = None
        return native.CorsairUnsubscribeFromEvents()

    def get_property(self, device_index, property_id):
        if (property_id & CorsairDevicePropertyType.Boolean) != 0:
            prop = c_bool()
            success = native.CorsairGetBoolPropertyValue(
                device_index, property_id, byref(prop))
            if success:
                return prop.value
        elif (property_id & CorsairDevicePropertyType.Int32) != 0:
            prop = c_int32()
            success = native.CorsairGetInt32PropertyValue(
                device_index, property_id, byref(prop))
            if success:
                return prop.value
        return None



import os
import platform
import re
import sys
from setuptools import setup


def read_version(filename='cuesdk/version.py'):
    """Parse a __version__ number from a source file"""
    with open(filename) as source:
        text = source.read()
        match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", text)
        if not match:
            msg = "Unable to find version number in {}".format(filename)
            raise RuntimeError(msg)
        version = match.group(1)
        return version


system = platform.system().lower()

if system not in ('windows', 'darwin'):
    msg = "{} system is not supported".format(system)
    raise RuntimeError(msg)


def package_files(directory):
    return [
        os.path.join('..', path, filename)
        for (path, directories, filenames) in os.walk(directory)
        for filename in filenames
    ]


setup(
    name="cuesdk",
    version=read_version(),
    packages=['cuesdk'],
    package_data={
        'cuesdk': package_files('cuesdk/bin'),
    },
    zip_safe=False,
    author="Corsair Memory, Inc.",
    license='MIT',
    url="https://github.com/CorsairOfficial/cue-sdk-python",
    description="Ctypes-based CUE SDK binding for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[],
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Games/Entertainment',
        'Topic :: System :: Hardware'
    ]
)



__all__ = ['ColorRgb']


class ColorRgb(object):
    def __init__(self, r: int = 0, g: int = 0, b: int = 0):
        def validate(c):
            if not 0 <= c <= 255:
                raise ValueError("The argument must be between 0 and 255.")
            return int(c)

        self.r = validate(r)
        self.g = validate(g)
        self.b = validate(b)

    @property
    def rgb(self):
        return (self.r, self.g, self.b)

    @property
    def hex(self):
        return "0x%06x" % (((self.r & 0xff) << 16) | ((self.g & 0xff) << 8) |
                           (self.b & 0xff))

    @staticmethod
    def _denorm_(fnum):
        return 255 * max(min(fnum, 1), 0)

    @classmethod
    def from_vec3(cls, *v3):
        dn = cls._denorm_
        col = (dn(c) for c in v3)
        return cls(*col)

    @classmethod
    def from_hexstr(cls, hexstr: str):
        val = hexstr.lstrip('#').lower().replace('0x', '').rjust(6, '0')
        return cls(*(int(val[i:i + 2], 16) for i in (0, 2, 4)))

    def __str__(self):
        return self.hex

    def __repr__(self):
        return "ColorRgb(%d, %d, %d)" % self.rgb

    def __iter__(self):
        return iter(self.rgb)
